# Copyright (C) 2014 Julian Metzler
# -*- coding: utf-8 -*-
# See the LICENSE file for the full license.

"""
Main protocol library
"""

import serial
import time

# Try importing the GPIO lib in case we're on a Raspberry Pi (to control the stop indicators)
try:
	import wiringpi
	HAVE_GPIO = True
except ImportError:
	HAVE_GPIO = False

"""
Example GPIO pinmap:

{
	0: 28,
	1: 29,
	2: 31,
	3: 30
}

(Display ID -> GPIO pin)
"""

class IBISMaster(object):
	def __init__(self, port, gpio_pinmap = {}):
		self.port = port
		self.gpio_pinmap = gpio_pinmap
		
		if HAVE_GPIO:
			self.gpio = wiringpi.GPIO(wiringpi.GPIO.WPI_MODE_GPIO)
			
			for pin in self.gpio_pinmap.values():
				self.gpio.pinMode(pin, self.gpio.OUTPUT)
		
		self.device = serial.serial_for_url(
			self.port,
			baudrate = 1200,
			bytesize = serial.SEVENBITS,
			parity = serial.PARITY_EVEN,
			stopbits = serial.STOPBITS_TWO
		)
	
	def hash(self, message):
		check_byte = 0x7F
		
		for char in message:
			byte = ord(char)
			check_byte = check_byte ^ byte
		
		message += chr(check_byte)
		return message
	
	def prepare_text(self, message):
		def _do_replace(message):
			message = message.replace(u"ä", "{")
			message = message.replace(u"ö", "|")
			message = message.replace(u"ü", "}")
			message = message.replace(u"ß", "~")
			message = message.replace(u"Ä", "[")
			message = message.replace(u"Ö", "\\")
			message = message.replace(u"Ü", "]")
			message = message.encode('utf-8')
			return message
		
		try:
			message = _do_replace(message)
		except UnicodeDecodeError:
			message = message.decode('utf-8')
			message = _do_replace(message)
		
		return message
	
	def set_stop_indicator(self, address, value):
		if not HAVE_GPIO:
			return False
		
		pin = self.gpio_pinmap.get(address, None)
		if pin is None:
			return False
		
		self.gpio.digitalWrite(pin, bool(value))
	
	def send_raw(self, data):
		#print repr(data)
		hex_data = ""
		for byte in data:
			hex_data += "<%s>" % hex(ord(byte))[2:].upper().rjust(2, "0")
		#print hex_data
		length = self.device.write(data)
		time.sleep(length * (12 / 1200.0))
		return length
	
	def send_message(self, message):
		message = self.hash(message + "\r")
		return self.send_raw(message)
	
	def send_line_number(self, line_number):
		message = "l%03i" % line_number
		return self.send_message(message)
	
	def send_special_character(self, character):
		message = "lE%02i" % character
		return self.send_message(message)
	
	def send_target_number(self, target_number):
		message = "z%03i" % target_number
		return self.send_message(message)
	
	def send_time(self, hours, minutes):
		message = "u%02i%02i" % (hours, minutes)
		return self.send_message(message)
	
	def send_date(self, day, month, year):
		message = "d%02i%02i%i" % (day, month, year)
		return self.send_message(message)
	
	def send_target_text__003a(self, text):
		text = self.prepare_text(text)
		blocks, remainder = divmod(len(text), 16)
		
		if remainder:
			blocks += 1
			text += " " * (16 - remainder)
		
		message = "zA%i%s" % (blocks, text.upper())
		return self.send_message(message)
	
	def send_target_text__021(self, text, id):
		text = self.prepare_text(text)
		blocks, remainder = divmod(len(text), 16)
		
		if remainder:
			blocks += 1
			text += " " * (16 - remainder)
		
		message = "aA%i%i%s" % (id, blocks, text.upper())
		return self.send_message(message)
	
	def send_next_stop__009(self, next_stop, length = 16):
		next_stop = self.prepare_text(next_stop)
		message = "v%s" % next_stop.upper().ljust(length)
		return self.send_message(message)
	
	def send_next_stop__003c(self, next_stop):
		next_stop = self.prepare_text(next_stop)
		blocks, remainder = divmod(len(next_stop), 4)
		
		if remainder:
			blocks += 1
			next_stop += " " * (4 - remainder)
		
		message = "zI%i%s" % (blocks, next_stop)
		return self.send_message(message)


# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

from .metadata import version as __version__
from .ibis_protocol import *
from .ibis_server import Server
from .ibis_client import Client
import ibis_simulation as simulation


# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Tool to graphically simulate the text display
(For now, it's specific to the kind of display I have, feel free to modify the font files to suit your needs)
"""

import json
import os
from PIL import Image, ImageDraw

class DisplayFontScanner(object):
	"""
	Class to scan the PNG files that make up the font and create a representation
	of the font that can be programmatically used.
	
	There must be a <dotspacing / 2> pixel border on all sides of the image.
	Each dot must be <dotsize> pixels in diameter and the spacing between adjacent dots
	must be <dotspacing> pixels.
	Inactive dots must be gray, active dots must be any other color.
	"""
	
	def __init__(self, dotsize = 47, dotspacing = 8):
		self.dotsize = dotsize
		self.dotspacing = dotspacing
	
	def get_real_coordinates(self, x, y):
		"""
		Calculate the real x and y pixel coordinates from x and y dot coordinates
		"""
		
		real_x = (self.dotspacing / 2) + x * self.dotspacing + (x + 1) * self.dotsize - (self.dotsize / 2)
		real_y = (self.dotspacing / 2) + y * self.dotspacing + (y + 1) * self.dotsize - (self.dotsize / 2)
		
		return real_x, real_y
		
	def get_dot(self, image, x, y):
		"""
		Get the state of the dot at the given coordinates
		image must be a PIL Image instance
		"""
		
		real_x, real_y = self.get_real_coordinates(x, y)
		r, g, b = image.getpixel((real_x, real_y))[:3]
		state = not r == g == b # Check if dot is gray
		
		return state
	
	def build_char_data(self, image):
		"""
		Build the character data from the given PIL Image instance
		"""
		
		width = divmod(image.size[0], (self.dotsize + self.dotspacing))[0]
		height = divmod(image.size[1], (self.dotsize + self.dotspacing))[0]
		
		dots = []
		for y in range(height):
			row = []
			for x in range(width):
				row.append(self.get_dot(image, x, y))
			dots.append(row)
		
		char_data = {
			'width': width,
			'height': height,
			'dots': dots,
		}
		
		return char_data
		
	def scan_file(self, filename):
		"""
		Scan a single PNG file and build the character data for it
		"""
		
		image = Image.open(filename)
		char_data = self.build_char_data(image)
		
		return char_data
	
	def scan_font(self, fontdir, outfile):
		"""
		Scan the font folder and build the font map
		"""
		
		fontmap = {}
		files = [os.path.join(fontdir, filename) for filename in os.listdir(fontdir)]
		
		for filename in files:
			char_name = os.path.basename(filename)[:-4]
			try:
				char_data = self.scan_file(filename)
			except:
				continue
			if char_name == "slash":
				char_name = "/"
			fontmap[char_name] = char_data
		
		with open(outfile, 'w') as f:
			json.dump(fontmap, f)
		
		return fontmap

class DisplayFontGenerator(object):
	"""
	Class to do the reverse of what the DisplayFontScanner is doing.
	Generates images from a fontmap.
	"""
	
	def __init__(self, dotsize = 47, dotspacing = 8):
		self.dotsize = dotsize
		self.dotspacing = dotspacing
	
	def get_real_coordinates(self, x, y):
		"""
		Calculate the real x and y pixel coordinates from x and y dot coordinates
		"""
		
		real_x = (self.dotspacing / 2) + x * self.dotspacing + (x + 1) * self.dotsize - (self.dotsize / 2)
		real_y = (self.dotspacing / 2) + y * self.dotspacing + (y + 1) * self.dotsize - (self.dotsize / 2)
		
		return real_x, real_y
	
	def get_bounding_box(self, x, y, size):
		"""
		Get the bounding box coordinates for drawing a circle with the diameter <size>
		"""
		
		x0 = x - (size / 2)
		y0 = y - (size / 2)
		x1 = x + (size / 2)
		y1 = y + (size / 2)
		
		return x0, y0, x1, y1
	
	def generate_image(self, char_data, outfile, inactive_color, active_color, bg_color):
		"""
		Generate a single image from a character map.
		"""
		
		width = char_data['width']
		height = char_data['height']
		dots = char_data['dots']
		
		im_width = width * (self.dotsize + self.dotspacing)
		im_height = height * (self.dotsize + self.dotspacing)
		
		image = Image.new("RGB", (im_width, im_height), bg_color)
		draw = ImageDraw.Draw(image)
		
		for y, row in enumerate(dots):
			for x, state in enumerate(row):
				color = active_color if state else inactive_color
				
				# Workaround because with dotsize 1 ellipses become invisible
				if self.dotsize > 1:
					center_x, center_y = self.get_real_coordinates(x, y)
					box = self.get_bounding_box(center_x, center_y, self.dotsize)
					draw.ellipse(box, fill = color)
				else:
					draw.point((x, y), fill = color)
		
		if type(outfile) is str:
			with open(outfile.encode('utf-8'), 'wb') as f:
				image.save(f, "PNG")
		else:
			image.save(outfile, "PNG")
	
	def generate_font(self, fontmap_file, outdir, inactive_color = (64, 64, 64), active_color = (192, 255, 0), bg_color = (0, 0, 0)):
		"""
		Generate all images for a given fontmap
		"""
		
		with open(fontmap_file, 'r') as f:
			fontmap = json.load(f)
		
		for char, char_data in fontmap.iteritems():
			name = ("slash" if char == "/" else char) + ".png"
			filename = os.path.join(outdir, name)
			self.generate_image(char_data, filename, inactive_color, active_color, bg_color)

class DisplayFont(object):
	"""
	This class represents a display font used to display text.
	"""
	
	def __init__(self, fontmap_file, spacing):
		self.fontmap_file = fontmap_file
		self.spacing = spacing
		with open(self.fontmap_file, 'r') as f:
			self.fontmap = json.load(f)
	
	def get_width(self, text):
		length = len(text)
		width = 0
		for index, char in enumerate(text):
			if char in self.fontmap.keys():
				char_data = self.fontmap[char]
				width += char_data['width']
				if index < length - 1:
					width += self.spacing
		
		return width
	
	def get_nominal_width(self, text = "ABCDEabcde12345"):
		"""
		Calculate the average character width, used for sorting fonts based on the
		space the characters take
		"""
		
		avg_width = self.get_width(text) / len(text)
		
		return avg_width
	
	def generate_text_data(self, text):
		"""
		Generate dot data for an image representing the given text
		"""
		
		text_dots = []
		
		for index, char in enumerate(text):
			if char not in self.fontmap.keys():
				continue
			
			char_data = self.fontmap[char]
			dots = char_data['dots']
			
			for y, row in enumerate(dots):
				try:
					text_dots[y] += row
				except IndexError:
					text_dots.append(row[:]) # WITHOUT THE [:] SHIT GOES TO HELL
				
				if index < len(text) - 1:
					text_dots[y] += [False] * self.spacing
		
		width = len(text_dots[0])
		height = len(text_dots)
		
		text_data = {
			'width': width,
			'height': height,
			'dots': text_dots,
		}
		
		return text_data

class DisplaySimulator(object):
	"""
	Class that manages DisplayFonts and simulates the graphical display of the real displays
	
	The 'fonts' parameter must list the fonts from widest to narrowest font
	"""
	
	def __init__(self, fonts, width = 120, height = 8):
		self.fonts = fonts
		self.width = width
		self.height = height
	
	def choose_font(self, text):
		"""
		Choose the font that best fills the available space for the given text
		"""
		
		for font in self.fonts:
			if font.get_width(text) <= self.width:
				return font
		
		return self.fonts[-1] # Return the narrowest font in case no font is small enough
	
	def generate_image(self, text, outfile, dotsize = 47, dotspacing = 8, inactive_color = (64, 64, 64), active_color = (192, 255, 0), bg_color = (0, 0, 0)):
		"""
		Generate an image representing the given text
		Returns an overflow flag and the dot data for analytical purposes
		"""
		
		font = self.choose_font(text)
		overflow = False
		
		while font.get_width(text) > self.width:
			overflow = True
			text = text[:-1]
		
		text_data = font.generate_text_data(text)
		
		for y, row in enumerate(text_data['dots']):
			if len(row) > self.width:
				row = row[:self.width]
			elif len(row) == self.width:
				pass
			else:
				diff = self.width - len(row)
				left = diff / 2
				right = diff - left
				text_data['dots'][y] = [False] * left + row + [False] * right
		
		text_data['width'] = self.width
		
		gen = DisplayFontGenerator(dotsize, dotspacing)
		gen.generate_image(text_data, outfile, inactive_color, active_color, bg_color)
		return overflow, text_data['dots']

def main():
	pass

if __name__ == "__main__":
	main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Client for the client-server system
"""

import argparse
import re
import socket
import time

from .ibis_utils import _receive_datagram, _send_datagram

class Client(object):
	def __init__(self, host, port = 4242, timeout = 5.0):
		self.host = host
		self.port = port
		self.timeout = timeout
		self.socket = None
	
	def send_raw_message(self, message, expect_reply = True):
		"""
		Send a message to the server
		"""
		
		reply = None
		
		try:
			sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			sock.settimeout(self.timeout)
			sock.connect((self.host, self.port))
			_send_datagram(sock, message)
			
			if expect_reply:
				reply = _receive_datagram(sock)
		finally:
			sock.close()
		
		return reply
	
	def set_enabled(self, address, value):
		"""
		Enable or disable a display
		"""
		
		return self.send_raw_message({'address': address, 'enable': value})
	
	def set_stop_indicator(self, address, state):
		"""
		Enable or disable a stop indicator
		"""
		
		return self.send_raw_message({'address': address, 'stop_indicator': state})
	
	def set_message(self, address, message, priority = 0, client = None):
		"""
		Set the message for a display
		"""
		
		message = {
			'address': address,
			'message': message,
			'priority': priority,
		}
		
		if client:
			message['client'] = client
		
		return self.send_raw_message(message)
	
	def make_text(self, text, duration = None):
		message = {'type': 'text', 'text': text}
		if duration:
			message['duration'] = duration
		
		return message
	
	def set_text(self, address, text, duration = None, priority = 0, client = None):
		"""
		Set a static text on a display
		"""
		
		return self.set_message(address, self.make_text(text, duration), priority, client)
	
	def make_time(self, format, duration = None):
		message = {'type': 'time', 'format': format}
		if duration:
			message['duration'] = duration
		
		return message
	
	def set_time(self, address, format, duration = None, priority = 0, client = None):
		"""
		Set a display to display the current time
		"""
		
		return self.set_message(address, self.make_time(format, duration), priority, client)
	
	def set_sequence(self, address, sequence, interval, priority = 0, client = None):
		"""
		Set a display to display a sequence of messages
		"""
		
		return self.set_message(address, {'type': 'sequence', 'messages': sequence, 'interval': interval}, priority, client)
	
	def get_current_text(self):
		"""
		Query the server for the text that is currently being displayed
		"""
		
		current = self.send_raw_message({'query': 'current_text'})
		current = dict([(int(key), value) for key, value in current.iteritems()])
		return current
	
	def get_buffer(self):
		"""
		Query the server for the text that is set to be displayed for each display
		"""
		
		buffer = self.send_raw_message({'query': 'buffer'})
		buffer = dict([(int(key), value) for key, value in buffer.iteritems()])
		return buffer
	
	def get_enabled(self):
		"""
		Query the server for the enabled states
		"""
		
		enabled = self.send_raw_message({'query': 'enabled'})
		enabled = dict([(int(key), value) for key, value in enabled.iteritems()])
		return enabled
	
	def get_stop_indicators(self):
		"""
		Query the server for the current state of the stop indicators
		"""
		
		indicators = self.send_raw_message({'query': 'stop_indicators'})
		indicators = dict([(int(key), value) for key, value in indicators.iteritems()])
		return indicators
	
	def get_all(self):
		"""
		Query the server for all available status information
		"""
		
		status = self.send_raw_message({'query': 'all'})
		for subsection, subdict in status.iteritems():
			subdict = dict([(int(key), value) for key, value in subdict.iteritems()])
			status[subsection] = subdict
		return status


# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

name = "pyIBIS"
version = "1.4.5"
description = "A Python library for controlling IBIS displays"
license = "AGPLv3"
author = "Julian Metzler"
author_email = "pyibis@mezgr.de"
requires = []
url = "https://github.com/Mezgrman/pyIBIS"
keywords = "led sign message board effect library wrapper serial text display ibis vdv300 bus next stop train"


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Server for the client-server system
"""

import argparse
import ibis
import json
import socket
import thread
import time

from .ibis_utils import _receive_datagram, _send_datagram

class Listener(object):
	def __init__(self, controller, port = 4245):
		self.controller = controller
		self.port = port
		self.running = False
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	
	def run(self):
		self.running = True
		
		# Open the network socket and listen
		self.socket.bind(('', self.port))
		print "Listening on port %i" % self.port
		self.socket.listen(1)
		
		try:
			while self.running:
				try:
					# Wait for someone to connect
					conn, addr = self.socket.accept()
					print "Accepted connection from %s on port %i" % addr
					
					# Load the datagram
					message = _receive_datagram(conn)
					
					if message is None:
						# We received an invalid datagram, just discard it
						continue
					
					success = True
					if 'enable' in message:
						try:
							if message['enable'] == 'toggle':
								state = not self.controller.get_enabled(int(message['address']))
							else:
								state = bool(message['enable'])
							
							success = self.controller.set_enabled(int(message['address']), state)
						except:
							success = False
						_send_datagram(conn, {'success': success})
					elif 'query' in message:
						if message['query'] == 'current_text':
							_send_datagram(conn, self.controller.current_text)
						elif message['query'] == 'buffer':
							_send_datagram(conn, self.controller.buffer)
						elif message['query'] == 'enabled':
							_send_datagram(conn, self.controller.enabled)
						elif message['query'] == 'stop_indicators':
							_send_datagram(conn, self.controller.stop_indicators)
						elif message['query'] == 'all':
							status = {
								'buffer': self.controller.buffer,
								'current_text': self.controller.current_text,
								'enabled': self.controller.enabled,
								'stop_indicators': self.controller.stop_indicators
							}
							_send_datagram(conn, status)
					elif 'stop_indicator' in message:
						try:
							if message['stop_indicator'] == 'toggle':
								state = not self.controller.stop_indicators[message['address']]
							else:
								state = bool(message['stop_indicator'])
							
							success = self.controller.set_stop_indicator(int(message['address']), state)
						except:
							success = False
						_send_datagram(conn, {'success': success})
					else:
						try:
							success = self.controller.set_message(message['address'], message['message'], priority = message.get('priority', 0), client = message.get('client', addr[0]))
						except:
							success = False
						_send_datagram(conn, {'success': success})
				except KeyboardInterrupt:
					self.quit()
		finally:
			self.socket.close()
	
	def quit(self):
		self.running = False

class Controller(object):
	DEBUG = False
	VERBOSE = True
	TIMEOUT = 120.0
	
	def __init__(self, master):
		self.master = master
		self.running = False
		
		self.buffer = {
			0: {
				'message': None,
				'priority': -1,
				'client': None,
				'current': -1,
				'last_refresh': 0.0,
				'last_update': 0.0
			},
			1: {
				'message': None,
				'priority': -1,
				'client': None,
				'current': -1,
				'last_refresh': 0.0,
				'last_update': 0.0
			},
			2: {
				'message': None,
				'priority': -1,
				'client': None,
				'current': -1,
				'last_refresh': 0.0,
				'last_update': 0.0
			},
			3: {
				'message': None,
				'priority': -1,
				'client': None,
				'current': -1,
				'last_refresh': 0.0,
				'last_update': 0.0
			}
		}
		
		self.enabled = {
			0: True,
			1: True, 
			2: True,
			3: True
		}
		
		self.current_text = {
			0: None,
			1: None,
			2: None,
			3: None
		}
		
		self.stop_indicators = {
			0: False,
			1: False,
			2: False,
			3: False
		}
		
		try:
			self.load_config()
		except:
			if self.VERBOSE:
				print "Failed to load configuration"
	
	def _reverse_prepare_text(self, message):
		def _do_replace(message):
			message = message.replace("{", u"ä")
			message = message.replace("|", u"ö")
			message = message.replace("}", u"ü")
			message = message.replace("~", u"ß")
			message = message.replace("[", u"Ä")
			message = message.replace("\\", u"Ö")
			message = message.replace("]", u"Ü")
			message = message.encode('utf-8')
			return message
		
		try:
			message = _do_replace(message)
		except UnicodeDecodeError:
			message = message.decode('utf-8')
			message = _do_replace(message)
		
		return message
	
	def save_config(self, filename = "ibis.json"):
		if self.VERBOSE:
			print "Saving configuration..."
		
		data = {
			'buffer': self.buffer,
			'current_text': self.current_text,
			'enabled': self.enabled,
			'stop_indicators': self.stop_indicators,
		}
		
		with open(filename, 'w') as f:
			f.write(json.dumps(data))
		
		if self.VERBOSE:
			print "Successfully saved configuration"
	
	def load_config(self, filename = "ibis.json"):
		if self.VERBOSE:
			print "Loading configuration..."
		
		with open(filename, 'r') as f:
			data = json.loads(f.read())
		
		for address, entry in data['buffer'].iteritems():
			if entry['message']:
				self.set_message(int(address), entry['message'], priority = entry.get('priority', 0), client = entry.get('client', None))
		
		for address, state in data['stop_indicators'].iteritems():
			self.set_stop_indicator(int(address), state)
		
		for address, state in data['enabled'].iteritems():
			self.set_enabled(int(address), state)
		
		if self.VERBOSE:
			print "Successfully loaded configuration"
	
	def set_stop_indicator(self, address, value):
		self.master.set_stop_indicator(address, value)
		self.stop_indicators[address] = value
		
		if self.VERBOSE:
			print "Stop indicator on display %i set to %s" % (address, str(value))
		
		self.save_config()
		return True
	
	def set_enabled(self, address, value):
		"""
		Enable or disable a display
		"""
		
		if address == -1:
			for i in range(4):
				self.set_enabled(i, value)
			return True
		
		self.enabled[address] = value
		
		if not self.enabled[address]:
			self.send_text(address, None) # This seems to fail quite often! Why?
		
		if self.VERBOSE:
			print "Power state of display %i changed to %s" % (address, str(value))
		
		self.save_config()
		return True
	
	def get_enabled(self, address):
		"""
		Return the enabled state for the given address and perform some logic
		in case we're dealing with address -1
		"""
		
		if address == -1:
			return self.enabled[0] and self.enabled[1] and self.enabled[2] and self.enabled[3]
		else:
			return self.enabled.get(address, False)
	
	def send_text(self, address, text):
		"""
		Send text to a display
		"""
		
		# Set the address on the multiplexer or send to all displays if address is -1
		if address == -1:
			for i in range(4):
				self.send_text(i, text)
			return
		elif address == 0:
			self.master.device.setDTR(0)
			self.master.device.setRTS(0)
		elif address == 1:
			self.master.device.setDTR(0)
			self.master.device.setRTS(1)
		elif address == 2:
			self.master.device.setDTR(1)
			self.master.device.setRTS(0)
		elif address == 3:
			self.master.device.setDTR(1)
			self.master.device.setRTS(1)
		
		# Truncate the text
		if text:
			text = text[:36]
		
		# Send the data
		self.master.send_next_stop__003c("" if text is None else text)
		if self.DEBUG:
			print address, text.encode('utf-8')
		
		# Save the current text
		self.current_text[address] = self._reverse_prepare_text(text).decode('utf-8') if text else None
	
	def set_message(self, address, message, priority = 0, client = None):
		"""
		Set the stuff to be displayed on a display, like a sequence of texts
		
		Message Examples
		
			Simple text
				{
					'type': 'text',
					'text': "Hello world"
				}
			
			Formatted time
				{
					'type': 'time',
					'format': "%d.%m.%Y %H:%M"
				}
			
			Two alternating texts with the same interval
				{
					'type': 'sequence',
					'messages': [
						{
							'type': 'time',
							'format': "%d.%m.%Y %H:%M"
						},
						{
							'type': 'text',
							'text': "SE50 Frankfurt"
						}
					],
					'interval': 5.0
				}
			
			Two alternating texts with different intervals
				{
					'type': 'sequence',
					'messages': [
						{
							'type': 'text',
							'text': "Next Stop",
							'duration': 2.0
						},
						{
							'type': 'text',
							'text': "Frankfurt",
							'duration': 5.0
						}
					],
					'interval': 5.0
				}
		
		Note: The 'interval' property of sequences is used for all messages that don't specify a duration of their own.
		"""
		
		def _filter_ascii(message):
			# Filter out everything that's not 7-bit ASCII
			if message['type'] == 'text':
				text = ""
				for char in message['text']:
					if ord(char) <= 127 or char in [u"ä", u"ö", u"ü", u"Ä", u"Ö", u"Ü", u"ß"]:
						text += char
				message['text'] = text
			elif message['type'] == 'time':
				text = ""
				for char in message['format']:
					if ord(char) <= 127 or char in [u"ä", u"ö", u"ü", u"Ä", u"Ö", u"Ü", u"ß"]:
						text += char
				message['format'] = text
			return message
		
		# Discard messages with a lower priority then the one in the buffer if not sent by the same client
		current_priority = self.buffer[address]['priority']
		current_client = self.buffer[address]['client']
		if priority < current_priority and client != current_client:
			if self.VERBOSE:
				print "Discarded message from %s for display %i (Priority was %i, current is %i set by %s)" % (client, address, priority, current_priority, current_client)
			return False
		
		if message['type'] == 'text':
			message = _filter_ascii(message)
		elif message['type'] == 'time':
			message = _filter_ascii(message)
		elif message['type'] == 'sequence':
			for index, msg in enumerate(message['messages']):
				message['messages'][index] = _filter_ascii(message['messages'][index])
		
		self.buffer[address]['message'] = message
		self.buffer[address]['priority'] = priority
		self.buffer[address]['client'] = client
		self.buffer[address]['current'] = -1
		self.buffer[address]['last_refresh'] = 0.0
		self.buffer[address]['last_update'] = 0.0
		
		if self.VERBOSE:
			print "Message on display %i set by %s with priority %i: %s" % (address, client, priority, str(message))
		
		self.save_config()
		
		return True
	
	def send_message(self, address, message):
		"""
		Send a single message of various types
		"""
		
		now = time.time()
		current_text = self.current_text[address]
		last_refresh = self.buffer[address]['last_refresh']
		last_update = self.buffer[address]['last_update']
		
		if message:
			if message['type'] == 'text':
				if current_text != message['text']:
					self.send_text(address, message['text'])
					self.buffer[address]['last_refresh'] = now
					self.buffer[address]['last_update'] = now
				elif last_refresh + self.TIMEOUT <= now:
					self.send_text(address, current_text)
					self.buffer[address]['last_refresh'] = now
			elif message['type'] == 'time':
				try:
					text = time.strftime(message['format'])
				except:
					text = time.strftime(message['format'].encode('utf-8'))
				if current_text != text:
					self.send_text(address, text)
					self.buffer[address]['last_refresh'] = now
					self.buffer[address]['last_update'] = now
				elif last_refresh + self.TIMEOUT <= now:
					self.send_text(address, current_text)
					self.buffer[address]['last_refresh'] = now
			elif message['type'] == 'sequence':
				default_interval = message['interval']
				messages = message['messages']
				current = self.buffer[address]['current']
				if current == -1 or current >= len(messages) - 1:
					next_message = 0
				else:
					next_message = current + 1
				duration = messages[current].get('duration', None)
				if duration is None:
					duration = default_interval
				if last_update + duration <= now:
					self.buffer[address]['current'] = next_message
					self.send_message(address, messages[next_message])
				elif last_refresh + self.TIMEOUT <= now:
					self.send_text(address, current_text)
					self.buffer[address]['last_refresh'] = now
		else:
			if current_text is not None:
				self.send_text(address, None)
				self.buffer[address]['last_update'] = now
	
	def selftest(self):
		self.send_text(-1, None)
		time.sleep(2)
		
		for i in range(4):
			self.set_stop_indicator(i, True)
		
		self.send_text(-1, "IBIS Server")
		time.sleep(2)
		self.send_text(-1, "by Mezgrman")
		time.sleep(2)
		self.send_text(-1, "www.mezgrman.de")
		time.sleep(2)
		
		for i in range(4):
			self.send_text(i, "Display %i" % i)
		
		time.sleep(5)
		self.send_text(-1, None)
		
		for i in range(4):
			self.set_stop_indicator(i, False)
	
	def process_buffer(self):
		"""
		Periodically check what's in the buffer and update the displays if necessary
		"""
		
		while self.running:
			for address in range(4):
				if self.enabled[address]:
					data = self.buffer[address]
					message = data['message']
					self.send_message(address, message)
			time.sleep(0.1)
	
	def run(self):
		self.running = True
		self.process_buffer()
	
	def quit(self):
		self.save_config()
		self.running = False

class Server(object):
	def __init__(self, serial_port, port = 4242, timeout = 120, gpio_pinmap = {}, verbose = False, debug = False, selftest = False):
		self.master = ibis.IBISMaster(serial_port, gpio_pinmap = gpio_pinmap)
		self.controller = Controller(self.master)
		self.controller.TIMEOUT = timeout
		self.controller.VERBOSE = verbose
		self.controller.DEBUG = debug
		
		if selftest:
			self.controller.selftest()
		
		self.listener = Listener(self.controller, port = port)
	
	def run(self):
		thread.start_new_thread(self.controller.run, ())
		self.listener.run()


# -*- coding: utf-8 -*-
# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Various convenience functions used internally in multiple places
"""

import json

def _receive_datagram(sock):
	# Receive and parse an incoming datagram (prefixed with its length)
	try:
		length = int(sock.recv(4))
		
		raw_data = ""
		read_length = 0
		while read_length < length:
			part = sock.recv(1024)
			raw_data += part
			read_length += len(part)
		
		datagram = json.loads(raw_data)
	except:
		return None
	return datagram

def _send_datagram(sock, data):
	# Build and send a datagram (prefixed with its length)
	raw_data = json.dumps(data)
	length = len(raw_data)
	datagram = "%04i%s" % (length, raw_data)
	sock.sendall(datagram)